/*
 *  kxen-win.c
 *  kxen
 *
 *  Copyright 2009 Citrix Systems, Inc. All rights reserved.
 *
 */

#include <err.h>
#include <stdint.h>

#include <kxenctllib.h>
#include <kxen_desc.h>
#include <kxen_info.h>

static HANDLE kxen_handle = INVALID_HANDLE_VALUE;

static void
close_and_unload(void)
{
    if (kxen_handle != INVALID_HANDLE_VALUE)
	kxen_close(kxen_handle);
    kxen_handle = INVALID_HANDLE_VALUE;

    (void)kxen_manage_driver(FALSE, FALSE);
}

int
kxen_setup(void)
{
    int ret;

    atexit(close_and_unload);

    kxen_handle = kxen_open(0, TRUE);
    if (kxen_handle == INVALID_HANDLE_VALUE)
	err(1, "kxen_open");

    /* Crank up the system clock. */
    timeBeginPeriod(1);

    ret = kxen_load(kxen_handle, "kxen.elf");
    if (ret && ret != EEXIST)
	err(1, "kxen_load");

    ret = kxen_init(kxen_handle);
    if (ret && ret != EEXIST)
	err(1, "kxen_init");

    return 0;
}

void
kxen_exit(void)
{
    (void)kxen_shutdown(kxen_handle);
}

struct kxen_vmap *
kxen_alloc_mem(size_t guest_size)
{
    int ret;
    struct kxen_vmappings_desc kvd;
    int nr_pages;
    struct kxen_vmap *vmappings;

    nr_pages = (guest_size + KXEN_PAGE_SIZE - 1) >> KXEN_PAGE_SHIFT;
    vmappings = malloc(sizeof(struct kxen_vmap) *
		       ((nr_pages + ((1 << KXEN_VMAPPING_SHIFT) - 1)) >>
			KXEN_VMAPPING_SHIFT));
    if (vmappings == NULL)
	return NULL;
    kvd.kvd_nrpages = nr_pages;
    kvd.kvd_vmappings = vmappings;
    ret = kxen_vmappings(kxen_handle, &kvd);
    if (ret)
	return NULL;
    return vmappings;
}

int
kxen_run(void)
{
    int ret;

    ret = kxen_execute(kxen_handle);
    return ret;
}

int
kxen_xc_hvmop(unsigned cmd, union kxen_hvmop_arg *kha)
{
    int ret;
    struct kxen_hvmop_desc khd;

    khd.khd_cmd = cmd;
    set_xen_guest_handle(khd.khd_kha, kha);

    ret = kxen_hvmop(kxen_handle, &khd);
    return ret;
}

int
kxen_xc_memop(unsigned cmd, union kxen_memop_arg *kma)
{
    int ret;
    struct kxen_memop_desc kmd;

    kmd.kmd_cmd = cmd;
    set_xen_guest_handle(kmd.kmd_kma, kma);

    ret = kxen_memop(kxen_handle, &kmd);
    return ret;
}

int
kxen_xc_domctl(struct xen_domctl *xd)
{
    int ret;
    struct kxen_domctl_desc kdd;

    xd->interface_version = XEN_DOMCTL_INTERFACE_VERSION;
    set_xen_guest_handle(kdd.kdd_xd, xd);

    ret = kxen_domctl(kxen_handle, &kdd);
    return ret;
}

int
kxen_xc_version(int cmd, void *arg)
{
    int ret;
    struct kxen_hypercall_desc khd;

    khd.khd_op = __HYPERVISOR_xen_version;
    khd.khd_arg[0] = cmd;
    khd.khd_arg[1] = (uint64_t)(uintptr_t)arg;

    ret = kxen_hypercall(kxen_handle, &khd);
    return ret;
}

void *
kxen_xc_mmapbatch(xen_pfn_t *pfn, int num, int prot, domid_t dom)
{
    int ret;
    struct kxen_mmapbatch_desc kmd;

    kmd.kmd_num = num;
    kmd.kmd_domid = dom;
    kmd.kmd_arr = pfn;
    ret = kxen_mmapbatch(kxen_handle, &kmd);
    if (ret != 0)
	kmd.kmd_addr = 0;
    return (void *)(uintptr_t)kmd.kmd_addr;
}

void
kxen_xc_munmap(void *addr, int num)
{
    struct kxen_munmap_desc kmd;

    kmd.kmd_num = num;
    kmd.kmd_addr = (uint64_t)(uintptr_t)addr;
    (void)kxen_munmap(kxen_handle, &kmd);
}

int
kxen_xc_keyhandler(const char *keys)
{
    int ret;

    ret = kxen_trigger_keyhandler(kxen_handle, keys);
    return ret;
}
