/* 
 *  kxen_ioctl.c
 *  kxen
 *
 *  Copyright 2009 Citrix Systems, Inc. All rights reserved.
 * 
 */

#include "kxen.h"

#include <ntddk.h>
#include <xen/errno.h>
#include <xen/types.h>

#include <kxen_ioctl.h>

#define KXEN_DEFINE_SYMBOLS_PROTO
#include <kxen_link.h>
KXEN_PROTOTYPES(extern)

#if 0
#define IOCTL_TRACE(fmt, ...) dprintk(fmt, __VA_ARGS__)
#else
#define IOCTL_TRACE(fmt, ...)
#endif

static int kxen_mode;
#define	KXEN_MODE_IDLE		0
#define	KXEN_MODE_LOADED	1
#define	KXEN_MODE_INITIALIZED	2

#define IOCTL_FAILURE(s, r, fmt, ...) do {	\
	IoStatus->Status = s;			\
	ret = r;				\
	fail_msg(fmt, __VA_ARGS__);		\
	ioctl_return = FALSE;			\
    } while (0)
    
#define SET_KXEN_MODE(m)			\
    kxen_mode = (m);
#define CHECK_KXEN_MODE(m, id) do {					\
	if (kxen_mode < (m)) {						\
	    IOCTL_FAILURE(STATUS_UNSUCCESSFUL, EINVAL,			\
			  "kxen_ioctl(" id ") invalid sequence");	\
	    goto out;							\
	}								\
    } while (0)
#define CHECK_KXEN_MODE_NOT(m, id) do {					\
	if (kxen_mode >= (m)) {						\
	    IOCTL_FAILURE(STATUS_UNSUCCESSFUL, EINVAL,			\
			  "kxen_ioctl(" id ") invalid sequence");	\
	    goto out;							\
	}								\
    } while (0)

#define KXEN_CALL(name, fn, arg, ...) do {				\
	IOCTL_TRACE("kxen_ioctl(" name ", %p, %x)\n", InputBuffer,	\
		    InputBufferLength);					\
	CHECK_KXEN_MODE(KXEN_MODE_INITIALIZED, name);			\
	if (InputBufferLength < sizeof(arg) || InputBuffer == NULL) {	\
	    IOCTL_FAILURE(STATUS_UNSUCCESSFUL, EINVAL,			\
			  "kxen_ioctl(" name ") input arguments");	\
	    goto out;							\
	}								\
	kxen_cpu_pin_dom0();						\
	try {								\
	    ret = fn((arg *)InputBuffer, __VA_ARGS__);			\
	} except(EXCEPTION_EXECUTE_HANDLER) {				\
	    IOCTL_FAILURE(GetExceptionCode(), EINVAL,			\
			  "kxen_ioctl(" name ") exception: 0x%08X",	\
			  IoStatus->Status);				\
	}								\
	kxen_cpu_unpin();						\
    } while (0)

BOOLEAN
kxen_ioctl(
    __in        FILE_OBJECT        *FileObject,
    __in        BOOLEAN             Wait,
    __in_opt    VOID               *InputBuffer,
    __in        ULONG               InputBufferLength,
    __out_opt   VOID               *OutputBuffer,
    __in        ULONG               OutputBufferLength,
    __in        ULONG               IoControlCode,
    __out       IO_STATUS_BLOCK    *IoStatus,
    __in        DEVICE_OBJECT      *DeviceObject
    )
{
    struct device_extension *devext;
    int func;
    int ret = EINVAL;
    BOOLEAN ioctl_return = TRUE;

    IoStatus->Status = STATUS_SUCCESS;
    IoStatus->Information = 0;

    devext = DeviceObject->DeviceExtension;

    func = FUNCTION_FROM_CTL_CODE(IoControlCode);

    if (func & KXEN_FLAG_INBUFFER) {
    }

    switch (IoControlCode) {
    case KXENVERSION:
	IOCTL_TRACE("kxen_ioctl(KXENVERSION, %p, %x)\n", OutputBuffer,
		    OutputBufferLength);
	if (OutputBufferLength < sizeof(struct kxen_version_desc) ||
	    OutputBuffer == NULL) {
	    IOCTL_FAILURE(STATUS_UNSUCCESSFUL, EINVAL,
			  "kxen_ioctl(KXENVERSION) output arguments");
	    break;
	}
	try {
	    ret = kxen_version((struct kxen_version_desc *)OutputBuffer);
	} except (EXCEPTION_EXECUTE_HANDLER) {
	    IOCTL_FAILURE(GetExceptionCode(), EINVAL,
			  "kxen_ioctl(KXENVERSION) exception: 0x%08X",
			  IoStatus->Status);
	}
	break;
    case KXENLOAD:
	IOCTL_TRACE("kxen_ioctl(KXENLOAD, %p, %x)\n", InputBuffer,
		    InputBufferLength);
	CHECK_KXEN_MODE_NOT(KXEN_MODE_LOADED, "KXENLOAD");
	try {
	    ret = kxen_load((struct kxen_load_desc *)InputBuffer);
	    if (ret == 0)
		SET_KXEN_MODE(KXEN_MODE_LOADED);
	} except (EXCEPTION_EXECUTE_HANDLER) {
	    IOCTL_FAILURE(GetExceptionCode(), EINVAL,
			  "kxen_ioctl(KXENLOAD) exception: 0x%08X",
			  IoStatus->Status);
	}
	break;
    case KXENUNLOAD:
	IOCTL_TRACE("kxen_ioctl(KXENUNLOAD)\n");
	CHECK_KXEN_MODE(KXEN_MODE_LOADED, "KXENUNLOAD");
	ret = kxen_unload();
	if (ret == 0)
	    SET_KXEN_MODE(KXEN_MODE_IDLE);
	break;
    case KXENINIT:
	IOCTL_TRACE("kxen_ioctl(KXENINIT)\n");
	CHECK_KXEN_MODE(KXEN_MODE_LOADED, "KXENINIT");
	CHECK_KXEN_MODE_NOT(KXEN_MODE_INITIALIZED, "KXENINIT");
	ret = kxen_init();
	if (ret == 0)
	    SET_KXEN_MODE(KXEN_MODE_INITIALIZED);
	/* XXX per VM/vcpu */
	KeInitializeEvent(&devext->de_runnable, NotificationEvent, FALSE);
	break;
    case KXENSHUTDOWN:
	IOCTL_TRACE("kxen_ioctl(KXENSHUTDOWN)\n");
	CHECK_KXEN_MODE(KXEN_MODE_INITIALIZED, "KXENSHUTDOWN");
	ret = kxen_shutdown();
	if (ret == 0)
	    kxen_mode = KXEN_MODE_LOADED;
	break;
    case KXENREBOOT:
	IOCTL_TRACE("kxen_ioctl(KXENREBOOT)\n");
	kdbgreboot();
	break;
    case KXENKEYHANDLER:
	IOCTL_TRACE("kxen_ioctl(KXENKEYHANDLER)\n");
	CHECK_KXEN_MODE(KXEN_MODE_INITIALIZED, "KXENKEYHANDLER");
	if (InputBufferLength == 0 || InputBuffer == NULL) {
	    IOCTL_FAILURE(STATUS_UNSUCCESSFUL, EINVAL,
			  "kxen_ioctl(KXENKEYHANDLER) input arguments");
	    break;
	}
	kxen_cpu_pin_dom0();
	try {
	    unsigned char *keys = (unsigned char *)InputBuffer;
	    unsigned int i;
	    for (i = 0; i < InputBufferLength; i++)
		kxen_handle_keypress(keys[i]);
	    ret = 0;
	} except(EXCEPTION_EXECUTE_HANDLER) {
	    IOCTL_FAILURE(GetExceptionCode(), EINVAL,
			  "kxen_ioctl(KXENKEYHANDLER) exception: 0x%08X",
			  IoStatus->Status);
	}
	kxen_cpu_unpin();
	break;

    case KXENHYPERCALL:
	KXEN_CALL("KXENHYPERCALL", kxen_do_hypercall,
		  struct kxen_hypercall_desc);
	break;
    case KXENMEMOP:
	KXEN_CALL("KXENMEMOP", kxen_do_memop, struct kxen_memop_desc);
	break;
    case KXENHVMOP:
	KXEN_CALL("KXENHVMOP", kxen_do_hvmop, struct kxen_hvmop_desc);
	break;
    case KXENDOMCTL:
	KXEN_CALL("KXENDOMCTL", kxen_do_domctl, struct kxen_domctl_desc);
	break;
    case KXENMMAPBATCH:
	KXEN_CALL("KXENMMAPBATCH", kxen_mmapbatch, struct kxen_mmapbatch_desc);
	break;
    case KXENMUNMAP:
	KXEN_CALL("KXENMUNMAP", kxen_munmap, struct kxen_munmap_desc);
	break;
    case KXENVMAPPINGS:
	KXEN_CALL("KXENVMAPPINGS", kxen_create_vmappings,
		  struct kxen_vmappings_desc, devext);
	break;
    case KXENEXECUTE:
	CHECK_KXEN_MODE(KXEN_MODE_INITIALIZED, "KXENEXECUTE");
	ret = kxen_execute();
	break;
    default:
	IOCTL_TRACE("kxen_ioctl(%lx)\n", IoControlCode);
	IOCTL_FAILURE(STATUS_NOT_IMPLEMENTED, EINVAL,
		      "kxen_ioctl(%lx) not implemented", IoControlCode);
	break;
    }

  out:
    IoStatus->Information = ret;
    return ioctl_return;
}
