/*
 *  kxen_load.c
 *  kxen
 *
 *  Copyright 2009 Citrix Systems, Inc. All rights reserved.
 *
 */

#include "kxen.h"

#include <ntddk.h>
#include <stdio.h>
#include <xen/errno.h>
#include <xen/types.h>

#include "kxen_ioctl.h"

#include <public/libelf.h>

#define KXEN_DEFINE_SYMBOLS_CODE
#include <kxen_link.h>
#define _ignore_
KXEN_PROTOTYPES( _ignore_ )
#undef _ignore_
KXEN_GET_SYMS(kxen_get_symbols)

#define snprintf _snprintf
#include <kxen_load.h>

int
kxen_load(struct kxen_load_desc *kld)
{
    int ret;
    uint8_t *image;
    struct elf_binary elf;
    const char *missing_symbol;

    if (kxen_hv) {
	fail_msg("load: already loaded");
	return EINVAL;
    }

    image = kld->kld_uvaddr;
    if (image == NULL) {
	fail_msg("load: no address");
	return EINVAL;
    }

    ret = elf_init(&elf, (const char *)image, kld->kld_size);
    if (ret != 0) {
	fail_msg("load: elf_init failed");
	ret = EINVAL;
	goto error;
    }

    elf_set_verbose(&elf);
    elf_parse_binary(&elf);

    kxen_size = (size_t)(elf.pend - elf.pstart);
    kxen_hv = kernel_malloc(kxen_size);
    if (kxen_hv == NULL) {
	fail_msg("load: malloc %d bytes failed", kxen_size);
	ret = ENOMEM;
	goto error;
    }

    elf.dest = (char *)kxen_hv;
    elf_load_binary(&elf);

    ret = elf_reloc(&elf);
    if (ret != 0) {
	fail_msg("load: elf_reloc failed");
	ret = EINVAL;
	goto error;
    }

    ret = kxen_get_symbols(&elf, kxen_hv, &missing_symbol);
    if (ret != 0) {
	fail_msg("load: kxen get symbol %s failed", missing_symbol);
	ret = EINVAL;
	goto error;
    }

    ret = kxen_fixup_os_code(&elf, kxen_hv, "windows");
    if (ret != 0) {
	fail_msg("load: fixup cpu code failed");
	ret = EINVAL;
	goto error;
    }

    dprintk("kxen_load done\n");
    ret = 0;
error:
    if (ret && kxen_hv) {
	kernel_free(kxen_hv, kxen_size);
	kxen_hv = NULL;
    }
    return ret;
}

void
kxen_unload_free_allocs(void)
{

    kxen_init_free_allocs();

    if (kxen_hv) {
	kernel_free(kxen_hv, kxen_size);
	kxen_hv = NULL;
    }
}

int
kxen_unload(void)
{

    if (kxen_hv == NULL)
	return 0;

    kxen_shutdown();

    kxen_unload_free_allocs();

    return 0;
}
