/*
 *  kxen_mem.c
 *  kxen
 *
 *  Copyright 2009 Citrix Systems, Inc. All rights reserved.
 *
 */

#include "kxen.h"

#include <ntddk.h>
#include <xen/errno.h>

#define KXEN_DEFINE_SYMBOLS_PROTO
#include <kxen_link.h>
KXEN_PROTOTYPES(extern)

void *
kernel_malloc(size_t size)
{
    return ExAllocatePoolWithTag(NonPagedPool, size, KXEN_POOL_TAG);
}

void
kernel_free(void *addr, size_t size)
{
    ExFreePoolWithTag(addr, KXEN_POOL_TAG);
}

uint32_t *
kernel_get_mfn_list(uintptr_t va, uint32_t npages)
{
    uint32_t *pages;
    MDL *mdl;
    PFN_NUMBER *mfnarray;
    uint32_t i;

    pages = kernel_malloc(npages * sizeof(uint32_t));
    if (pages == NULL)
	return NULL;

    mdl = IoAllocateMdl((void *)va, npages << PAGE_SHIFT, FALSE, TRUE, NULL);
    if (mdl == NULL)
	goto error;

    MmBuildMdlForNonPagedPool(mdl);

    mfnarray = MmGetMdlPfnArray(mdl);
    for (i = 0; i < npages; i++) {
        /* N.B. mfns are stored at 32 bit entries allowing for a maximum
           memory size of 16TB. */
	pages[i] = (uint32_t)mfnarray[i];
	if (pages[i] > kxen_info->ki_max_page) {
	    fail_msg("kernel_get_mfn_list: %x > max_page\n", pages[i],
		     kxen_info->ki_max_page);
	    goto error;
	}
    }

  out:
    if (mdl)
	IoFreeMdl(mdl);
    return pages;

  error:
    if (pages)
	kernel_free(pages, npages * sizeof(uint32_t));
    pages = NULL;
    goto out;
}

#define MAX_MDL_MAP_PAGES ((65536 - 1 - sizeof(MDL)) / sizeof(PFN_NUMBER))

static void *
mmap_pages(unsigned int num, xen_pfn_t *mfns)
{
    MDL *mdl;
    PFN_NUMBER *pfn;
    void *addr;
    unsigned int i;

    mdl = ExAllocatePoolWithTag(NonPagedPool, sizeof(MDL) +
				sizeof(PFN_NUMBER) * num,
				KXEN_POOL_TAG);
    if (mdl == NULL) {
	fail_msg("kxen_map_pages_to_user: alloc for mdl failed %x\n");
	return NULL;
    }
    memset(mdl, 0, sizeof(*mdl));

    mdl->ByteCount = num << PAGE_SHIFT;

    pfn = MmGetMdlPfnArray(mdl);
    for (i = 0; i < num; i++)
	pfn[i] = mfns[i];

    try {
	addr = MmMapLockedPagesSpecifyCache(mdl, UserMode, MmCached, NULL,
					    FALSE, LowPagePriority);
    } except (EXCEPTION_EXECUTE_HANDLER) {
	fail_msg("kxen_map_pages_to_user: "
		 "MmMapLockedPagesSpecifyCache failed %x\n",
		 GetExceptionCode());
	addr = NULL;
    }

    ExFreePoolWithTag(mdl, KXEN_POOL_TAG);

    return addr;
}

int
kxen_mmapbatch(struct kxen_mmapbatch_desc *kmmapbd)
{
    void *addr;
    int ret;
    struct kxen_memop_desc kmemopd;
    union kxen_memop_arg kmemopa;

    if (kmmapbd->kmd_num > MAX_MDL_MAP_PAGES) {
	fail_msg("kxen_map_pages_to_user: %d > MAX_MDL_MAP_PAGES %d\n",
		 kmmapbd->kmd_num, MAX_MDL_MAP_PAGES);
	return EINVAL;
    }

    kmemopa.translate_gpfn.domid = kmmapbd->kmd_domid;
    kmemopa.translate_gpfn.nr_gpfns = kmmapbd->kmd_num;
    set_xen_guest_handle(kmemopa.translate_gpfn.gpfn_list,
			 kmmapbd->kmd_arr);
    set_xen_guest_handle(kmemopa.translate_gpfn.mfn_list,
			 kmmapbd->kmd_arr);
    kmemopd.kmd_cmd = XENMEM_translate_gpfn_list;
    set_xen_guest_handle(kmemopd.kmd_kma, &kmemopa);
    ret = kxen_do_memop(&kmemopd);
    if (ret != 0) {
	fail_msg("kxen_map_pages_to_user: "
		 "XENMEM_translate_gpfn_list failed %x\n", ret);
	return ret;
    }

    addr = mmap_pages(kmmapbd->kmd_num, kmmapbd->kmd_arr);
    if (addr)
	kmmapbd->kmd_addr = (uint64_t)(uintptr_t)addr;
    else
	ret = EINVAL;

    return ret;
}

static int
munmap_pages(unsigned int num, uint8_t *addr)
{
    MDL *mdl;
    unsigned int i;
    int ret;
    PFN_NUMBER *pfn;
    PHYSICAL_ADDRESS pa;

    mdl = ExAllocatePoolWithTag(NonPagedPool,
				sizeof(MDL) + sizeof(PFN_NUMBER) * num,
				KXEN_POOL_TAG);
    if (mdl == NULL) {
	fail_msg("kxen_unmap_pages_from_user: alloc for mdl failed\n");
	return EINVAL;
    }

    memset(mdl, 0, sizeof(*mdl));

    mdl->ByteCount = num << PAGE_SHIFT;
    mdl->MdlFlags = MDL_PAGES_LOCKED;

    pfn = MmGetMdlPfnArray(mdl);
    for (i = 0; i < num; i++) {
	pa = MmGetPhysicalAddress(addr + (i << PAGE_SHIFT));
	pfn[i] = (PFN_NUMBER)(pa.QuadPart >> PAGE_SHIFT);
    }

    try {
	MmUnmapLockedPages(addr, mdl);
	ret = 0;
    } except (EXCEPTION_EXECUTE_HANDLER) {
	fail_msg("kxen_unmap_pages_from_user: "
		 "MmUnmapLockedPages failed %x\n", GetExceptionCode());
	ret = EINVAL;
    }

    ExFreePoolWithTag(mdl, KXEN_POOL_TAG);

    return ret;
}

int
kxen_munmap(struct kxen_munmap_desc *kmd)
{

    kxen_printk("kxen WARN: kxen_unmap_pages_from_user(%p,%x)\n",
		kmd->kmd_addr, kmd->kmd_num);

    return munmap_pages(kmd->kmd_num, (uint8_t *)kmd->kmd_addr);
}

int
kxen_create_vmappings(struct kxen_vmappings_desc *kvd,
		      struct device_extension *devext)
{
    struct kxen_vmap *vmappings, *vmaptable;
    PHYSICAL_ADDRESS lowAddress;
    PHYSICAL_ADDRESS highAddress;
    PHYSICAL_ADDRESS skipBytes;
    int ret = 0;
    int i, j, n;
    uint8_t *va;
    PFN_NUMBER *pfns;

    if (devext->de_nrpages) {
	fail_msg("kxen_create_vmappings: cannot change nr pages");
	return EINVAL;
    }

    devext->de_nrpages = kvd->kvd_nrpages;

    devext->de_vm_info.vi_p2m = kernel_malloc(devext->de_nrpages *
					      sizeof(uint32_t));
    if (devext->de_vm_info.vi_p2m == NULL) {
	fail_msg("kxen_create_vmappings: alloc p2m failed");
	ret = ENOMEM;
	goto out;
    }
    memset(devext->de_vm_info.vi_p2m, 0, devext->de_nrpages * sizeof(uint32_t));

#define VMAPPING_MDL_PAGES (1 << KXEN_VMAPPING_SHIFT)

    devext->de_nrmdls = (devext->de_nrpages + VMAPPING_MDL_PAGES - 1) >>
	KXEN_VMAPPING_SHIFT;

    devext->de_mdls = (MDL **)kernel_malloc(devext->de_nrmdls * sizeof(MDL *));
    if (devext->de_mdls == NULL) {
	fail_msg("kxen_create_vmappings: alloc mdl array failed");
	ret = ENOMEM;
	goto out;
    }
    memset(devext->de_mdls, 0, devext->de_nrmdls * sizeof(MDL *));

    vmappings = kvd->kvd_vmappings;

    vmaptable = (struct kxen_vmap *)kxen_info->ki_vmaptable;

    lowAddress.QuadPart = 0;
    highAddress.QuadPart = 0xfffffffffffff000UI64;
    skipBytes.QuadPart = 0;

    for (i = 0; i < devext->de_nrmdls; i++) {
	n = devext->de_nrpages - (i << KXEN_VMAPPING_SHIFT);
	if (n > VMAPPING_MDL_PAGES)
	    n = VMAPPING_MDL_PAGES;
	devext->de_mdls[i] = MmAllocatePagesForMdl(lowAddress, highAddress,
						   skipBytes, n << PAGE_SHIFT);
	if (devext->de_mdls[i] == NULL ||
	    devext->de_mdls[i]->ByteCount != n << PAGE_SHIFT) {
	    fail_msg("kxen_create_vmappings: alloc mdl failed");
	    ret = ENOMEM;
	    goto out;
	}
	try {
	    va = MmMapLockedPagesSpecifyCache(devext->de_mdls[i], UserMode,
					      MmCached, NULL, FALSE,
					      LowPagePriority);
	} except (EXCEPTION_EXECUTE_HANDLER) {
	    fail_msg("kxen_create_vmappings: failed to map pages to userspace");
	    ret = EINVAL;
	}
	if (ret)
	    goto out;

	try {
	    vmappings[i].vaddr = va;
	} except (EXCEPTION_EXECUTE_HANDLER) {
	    fail_msg("kxen_create_vmappings: update vmappings failed");
	    ret = EINVAL;
	}
	if (ret)
	    goto out;

	pfns = MmGetMdlPfnArray(devext->de_mdls[i]);
	for (j = 0; j < n; j++) {
            /* N.B. pfns are stored at 32 bit values so must be below 16TB */
	    ASSERT(pfns[j] < kxen_info->ki_max_page);
	    devext->de_vm_info.vi_p2m[(i << KXEN_VMAPPING_SHIFT) + j] =
                (uint32_t)pfns[j];
	    vmaptable[pfns[j]].vaddr = (va + (j << PAGE_SHIFT));
	}
    }

    devext->de_vm_info.vi_nrpages = devext->de_nrpages;

  out:
    if (ret)
	kxen_free_vmappings(devext);
    return ret;
}

void
kxen_free_vmappings(struct device_extension *devext)
{
    int i;

    if (devext->de_mdls) {
	for (i = 0; i < devext->de_nrmdls; i++) {
	    if (devext->de_mdls[i] == NULL)
		break;
	    MmFreePagesFromMdl(devext->de_mdls[i]);
	    ExFreePool(devext->de_mdls[i]);
	}
	kernel_free(devext->de_mdls, devext->de_nrmdls * sizeof(MDL *));
	devext->de_mdls = NULL;
    }

    if (devext->de_vm_info.vi_p2m)
	kernel_free(devext->de_vm_info.vi_p2m,
		    devext->de_nrpages * sizeof(uint32_t));
    devext->de_vm_info.vi_p2m = NULL;

    devext->de_nrpages = 0;
}

void * __cdecl
kxen_map_pages(unsigned int num, xen_pfn_t *mfns)
{

    return mmap_pages(num, mfns);
}

int __cdecl
kxen_unmap_pages(unsigned int num, void *va)
{

    return munmap_pages(num, va);
}
