
#include "kxen.h"
#include <kxen_ioctl.h>

#define KXEN_DEFINE_SYMBOLS_PROTO
#include <kxen_link.h>
KXEN_PROTOTYPES(extern)

#define KXEN_HEAP_SIZE (16 * 1024 * 1024)

static uint32_t *xenheap_pages = NULL;
static void *vmaptable = NULL;
static unsigned int vmaptable_size;
static void *frametable = NULL;
static unsigned int frametable_size;
static unsigned char *xenheap_va = NULL;

static KDPC kxen_cpu_ipi_dpc[MAXIMUM_PROCESSORS];
static PETHREAD kxen_idle_thread[MAXIMUM_PROCESSORS];
static KEVENT kxen_idle_thread_event[MAXIMUM_PROCESSORS];
static KTIMER kxen_cpu_timer[MAXIMUM_PROCESSORS];
static KDPC kxen_cpu_timer_dpc[MAXIMUM_PROCESSORS];
BOOLEAN kxen_shutting_down = FALSE;

static LARGE_INTEGER kxen_ready_time;
static BOOLEAN kxen_is_ready = FALSE;

extern BOOLEAN *KdDebuggerEnabled;

int
kxen_except_handler(unsigned int code, struct _EXCEPTION_POINTERS *ep)
{
    kxen_info->ki_running = 0;
    kxen_shutting_down = TRUE;
    dprintk("ldkxen 0x%p; gdb kxen\n", kxen_hv);
    dprintk(".cxr 0x%p\n", ep->ContextRecord);
    if (*KdDebuggerEnabled)
	DbgBreakPoint();
    return EXCEPTION_EXECUTE_HANDLER;
}

static void
kxen_cpu_ipi_cb(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
    unsigned int vector = (unsigned int)(ULONG_PTR)SystemArgument1;
    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(DeferredContext);
    UNREFERENCED_PARAMETER(SystemArgument2);

    if (vector)
	kxen_dispatch_ipi(vector);
}

static void __cdecl
kxen_cpu_ipi(unsigned int host_cpu, unsigned int vector)
{
    PKDPC dpc;

    if (host_cpu < MAXIMUM_PROCESSORS) {
        dpc = &kxen_cpu_ipi_dpc[host_cpu];
        KeInsertQueueDpc(dpc, (PVOID)(ULONG_PTR)vector, NULL);
    }
}

void
kxen_idle_thread_fn(void *context)
{
    unsigned int host_cpu = (unsigned int)(ULONG_PTR)context;

    KeSetSystemAffinityThread((ULONG_PTR)1 << host_cpu);
    dprintk("kxen_idle_thread on cpu %d\n", host_cpu);

    while (kxen_shutting_down == FALSE) {
	(void)KeWaitForSingleObject(&kxen_idle_thread_event[host_cpu],
				    Executive, KernelMode, TRUE, NULL);
	KeClearEvent(&kxen_idle_thread_event[host_cpu]);
	if (kxen_info->ki_running)
	    kxen_do_softirq();
    }
}

static void __cdecl
kxen_signal_idle_thread(unsigned int host_cpu)
{

    KeSetEvent(&kxen_idle_thread_event[host_cpu], 0, FALSE);
}

static void __cdecl
kxen_set_host_preemption(unsigned int disable)
{
    static KIRQL oldIrql[MAXIMUM_PROCESSORS];

    if (disable) {
	KeRaiseIrql(DISPATCH_LEVEL, &oldIrql[KeGetCurrentProcessorNumber()]);
    } else {
	KeLowerIrql(oldIrql[KeGetCurrentProcessorNumber()]);
    }
}

static void
kxen_cpu_timer_cb(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
    unsigned int host_cpu = (unsigned int)(ULONG_PTR)DeferredContext;
    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(SystemArgument1);
    UNREFERENCED_PARAMETER(SystemArgument2);

    if (kxen_info && kxen_info->ki_running && kxen_info->ki_timer_interrupt)
	kxen_info->ki_timer_interrupt();
}

static void __cdecl
kxen_set_cpu_timer(unsigned int host_cpu, uint64_t expire)
{
    LARGE_INTEGER timeDue;

    if (host_cpu >= MAXIMUM_PROCESSORS || !kxen_idle_thread[host_cpu])
	return;
    timeDue.QuadPart = TIME_RELATIVE(TIME_NANO(expire));
    KeSetTimerEx(&kxen_cpu_timer[host_cpu], timeDue, 0,
		 &kxen_cpu_timer_dpc[host_cpu]);
}

static uint64_t __cdecl
kxen_get_hostime(void)
{
    LARGE_INTEGER time;

    KeQuerySystemTime(&time);

    return (uint64_t)time.QuadPart;
}

static uint32_t __cdecl
kxen_get_unixtime(void)
{
    LARGE_INTEGER filetime;

    KeQuerySystemTime(&filetime);

    return (uint32_t)((filetime.QuadPart - 116444736000000000LL) / 10000000);
}

static void __cdecl
kxen_halt_vm(struct vm_info *vm_info)
{
    struct device_extension *devext;

    devext = CONTAINING_RECORD(vm_info, struct device_extension, de_vm_info);

    (void)KeWaitForSingleObject(&devext->de_runnable, Executive, KernelMode,
				TRUE, NULL);
    KeClearEvent(&devext->de_runnable);
}

static void __cdecl
kxen_wake_vm(struct vm_info *vm_info)
{
    struct device_extension *devext;

    devext = CONTAINING_RECORD(vm_info, struct device_extension, de_vm_info);

    KeSetEvent(&devext->de_runnable, 0, FALSE);
}

void
kxen_init_free_allocs(void)
{
    if (vmaptable) {
	kernel_free(vmaptable, vmaptable_size);
	vmaptable = NULL;
    }
    if (frametable) {
	kernel_free(frametable, frametable_size);
	frametable = NULL;
    }
    if (xenheap_pages) {
	kernel_free(xenheap_pages, KXEN_HEAP_SIZE >> PAGE_SHIFT);
	xenheap_pages = NULL;
    }
    if (xenheap_va) {
	kernel_free(xenheap_va, KXEN_HEAP_SIZE);
	xenheap_va = NULL;
    }
}

static NTSTATUS
init_cpu_dpc(KDPC *dpc, unsigned int host_cpu,
	     void (*cb)(KDPC *, void *, void *, void *), void *arg)
{
    KeInitializeDpc(dpc, cb, arg);
    KeSetTargetProcessorDpc(dpc, (CCHAR)host_cpu);
    KeSetImportanceDpc(dpc, HighImportance);

    return STATUS_SUCCESS;
}

int
kxen_init(void)
{
    int ret = -1;
    KAFFINITY affinity;
    unsigned int host_cpu;

    kxen_shutting_down = FALSE;

    dprintk("vvvvvvvvvvvvvvvvv\n"
	    "ldkxen %p; gdb kxen\n"
	    "^^^^^^^^^^^^^^^^^\n", kxen_hv);

    kxen_info->ki_printf = kxen_dprintk;
    kxen_info->ki_max_page = MAX_MACHINE_FRAME;
    kxen_info->ki_kick_cpu = kxen_cpu_ipi;
    kxen_info->ki_signal_idle_thread = kxen_signal_idle_thread;
    kxen_info->ki_set_host_preemption = kxen_set_host_preemption;
    kxen_info->ki_set_timer = kxen_set_cpu_timer;
    kxen_info->ki_get_hostime = kxen_get_hostime;
    kxen_info->ki_get_unixtime = kxen_get_unixtime;
    kxen_info->ki_cpu_active_mask = 0;
    kxen_info->ki_hostime_tick_frequency = KXEN_HOST_TIME_FREQUENCY;
    kxen_info->ki_halt_vm = kxen_halt_vm;
    kxen_info->ki_wake_vm = kxen_wake_vm;
    kxen_info->ki_on_each_cpu = kxen_on_each_cpu;
    kxen_info->ki_map_pages = kxen_map_pages;
    kxen_info->ki_unmap_pages = kxen_unmap_pages;

    kxen_set_cpu_active_mask(&kxen_info->ki_cpu_active_mask,
			     sizeof(kxen_info->ki_cpu_active_mask));

    xenheap_va = kernel_malloc(KXEN_HEAP_SIZE);
    if (xenheap_va == NULL)
	goto out;
    kxen_info->ki_xenheap_va = xenheap_va;

    kxen_info->ki_xenheap_page_count = KXEN_HEAP_SIZE >> PAGE_SHIFT;
    xenheap_pages = kernel_get_mfn_list((uintptr_t)xenheap_va,
					kxen_info->ki_xenheap_page_count);
    if (xenheap_pages == NULL)
	goto out;
    kxen_info->ki_xenheap_pages = xenheap_pages;

    vmaptable_size = kxen_info->ki_max_page * sizeof(struct kxen_vmap);
    vmaptable_size = ((vmaptable_size + PAGE_SIZE-1) & ~(PAGE_SIZE-1));
    vmaptable = kernel_malloc(vmaptable_size);
    if (vmaptable == NULL)
	goto out;
    memset(vmaptable, 0, vmaptable_size);
    kxen_info->ki_vmaptable = vmaptable;

    frametable_size = kxen_info->ki_max_page * kxen_sizeof_struct_page_info;
    frametable_size = ((frametable_size + PAGE_SIZE-1) & ~(PAGE_SIZE-1));
    frametable = kernel_malloc(frametable_size);
    if (frametable == NULL)
	goto out;
    kxen_info->ki_frametable = frametable;

    affinity = KeQueryActiveProcessors();
    for (host_cpu = 0; host_cpu < MAXIMUM_PROCESSORS; host_cpu++) {
	PKDPC dpc;
	HANDLE handle;
	NTSTATUS status;
	    
	kxen_idle_thread[host_cpu] = NULL;

	if ((affinity & ((ULONG_PTR)1 << host_cpu)) == 0)
	    continue;
	init_cpu_dpc(&kxen_cpu_ipi_dpc[host_cpu], host_cpu,
		     kxen_cpu_ipi_cb, NULL);

	KeInitializeEvent(&kxen_idle_thread_event[host_cpu],
			  NotificationEvent, FALSE);
	status = PsCreateSystemThread(&handle, 0, NULL, NULL, NULL,
				      kxen_idle_thread_fn,
                                      (PVOID)(ULONG_PTR)host_cpu);
	if (!NT_SUCCESS(status)) {
	    fail_msg("create idle thread failed: %x", status);
	    break;
	}

	status = ObReferenceObjectByHandle(handle, THREAD_ALL_ACCESS, NULL,
					   KernelMode,
					   &kxen_idle_thread[host_cpu], NULL);
	ZwClose(handle);
	if (!NT_SUCCESS(status)) {
	    fail_msg("get reference to idle thread %d faile: %x", status);
	    break;
	}

	KeInitializeTimer(&kxen_cpu_timer[host_cpu]);

	init_cpu_dpc(&kxen_cpu_timer_dpc[host_cpu], host_cpu,
		     kxen_cpu_timer_cb, (void *)(ULONG_PTR)host_cpu);
    }

    kxen_info->ki_running = 1;

    kxen_cpu_pin_first();
    try {
	kxen_start_xen();
	ret = 0;
    } except (kxen_except_handler(GetExceptionCode(),
				  GetExceptionInformation())) {
	ret = -1;
    }
    kxen_cpu_unpin();

    KeQuerySystemTime(&kxen_ready_time);
    kxen_ready_time.QuadPart += TIME_MS(2500);

out:
    if (ret)
	kxen_init_free_allocs();
    return ret;
}

int
kxen_shutdown(void)
{
    unsigned int host_cpu;

    if (kxen_info)
	kxen_info->ki_running = 0;
    kxen_shutting_down = TRUE;

    for (host_cpu = 0; host_cpu < MAXIMUM_PROCESSORS; host_cpu++) {
	if (kxen_idle_thread[host_cpu] == NULL)
	    continue;
	KeSetEvent(&kxen_idle_thread_event[host_cpu], 0, FALSE);
	KeCancelTimer(&kxen_cpu_timer[host_cpu]);
        KeWaitForSingleObject(kxen_idle_thread[host_cpu], Executive, KernelMode,
                              FALSE, NULL);
        ObDereferenceObject(kxen_idle_thread[host_cpu]);
        kxen_idle_thread[host_cpu] = NULL;
    }
    KeFlushQueuedDpcs();

    /* XXX per VM */
    if (kxen_devext)
	kxen_free_vmappings(kxen_devext);

    return 0;
}

int
kxen_version(struct kxen_version_desc *kvd)
{

    kvd->kvd_driver_version_major = KXEN_DRIVER_VERSION_MAJOR;
    kvd->kvd_driver_version_minor = KXEN_DRIVER_VERSION_MINOR;
    memset(kvd->kvd_driver_version_tag, 0, sizeof(kvd->kvd_driver_version_tag));
    strncpy(kvd->kvd_driver_version_tag, KXEN_DRIVER_VERSION_TAG,
	    sizeof(kvd->kvd_driver_version_tag));
    
    return 0;
}

int
kxen_execute(void)
{
    int ret;
    struct vm_info *vi;

    if (kxen_is_ready == FALSE) {
	KeDelayExecutionThread(KernelMode, FALSE, &kxen_ready_time);
	kxen_is_ready = TRUE;
    }

    vi = &kxen_devext->de_vm_info;

    kxen_cpu_pin_vm();
    try {
	ret = kxen_run_vm(vi);
    } except (kxen_except_handler(GetExceptionCode(),
				  GetExceptionInformation())) {
	ret = -1;
    }
    kxen_cpu_unpin();

    return ret;
}
